{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
{-# OPTIONS -fgenerics #-}
{-# OPTIONS -fth #-}

module NominalAbs where

import SYBnew
import NominalBase
import NominalSubst

import List((\\),union)

class Nominal a => BType a where
    bound :: a -> [Name]
    equiv :: a -> a -> Maybe Perm

data Abs a b = Abs a b deriving Show

$(derive [''Abs])

instance (BType a,  Nominal b) => Nominal (Abs a b) where
  swap t (Abs a b)            =  Abs (swap t a) (swap t b)
  fresh a (Abs x y)           =  a `elem` (bound x) || fresh a y
  aeq (Abs x y) (Abs x' y')  =  x `aeq` x' && y `aeq` y' 
				  || case equiv x x' 
                                     of  Just pi ->  all (\a -> fresh a y') (bound x)
                                                     && aeq y (perm pi y')
                                         Nothing ->  False

unAbs :: (FreshMonad m, BType a, Nominal a, Nominal b) => Abs a b -> m (a,b)
unAbs (Abs x y) = let bs = bound x in
		    do bs' <- mapM renameFM bs
		       return (perm (zip bs bs') (x,y))

instance (BType t, Nominal t, Nominal u,Subst a t,Subst a u) => Subst a (Abs t u) where
    subst a t abs = do (x,y) <- unAbs abs; 
		       x' <- subst a t x;
		       y' <- subst a t y;
		       return (Abs x y')



instance (BType t, Nominal t, Nominal u,FreeVars a t,FreeVars a u) => FreeVars a (Abs t u) where
    fvs dummy (Abs a x) = (fvs dummy a `union` fvs dummy x) \\ bound a




instance BType Name where
    bound a = [a]
    equiv a b = Just [(a,b)]


data Binding n a = Binding n a deriving (Show, Eq)

instance (Nominal a, Nominal b) => Nominal (Binding a b) where

instance (BType n, Eq a,Nominal a) => BType (Binding n a) where
    bound (Binding a x) = bound a
    equiv (Binding a x) (Binding b y) = if x == y then equiv a b else Nothing

   


