{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
{-# OPTIONS -fgenerics #-}
{-# OPTIONS -fth #-}

{- Nominal library base
   James Cheney
-}

module NominalBase

where

import SYBnew
import Data.Generics(Unit(Unit),(:+:)(Inl,Inr),(:*:)((:*:)))


data Name = Name String (Maybe Int) deriving Eq
type Trans = (Name,Name)
type Perm = [Trans]

$(derive [''Name])

instance Show Name where
    showsPrec i (Name s (Just k)) = 
	showString s . showString "$" . showsPrec i k
    showsPrec i (Name s Nothing) = 
	showString s


class Nominal a where
    swap                         :: Trans -> a -> a
    swap {|Unit|}    t Unit      =  Unit
    swap {|a :+: b|} t (Inl x)   =  Inl (swap t x)
    swap {|a :+: b|} t (Inr x)   =  Inr (swap t x)
    swap {|a :*: b|} t (x :*: y) =  swap t x :*: swap t y
				   
    fresh                         :: Name -> a -> Bool
    fresh {|Unit|}   a  Unit      =  True
    fresh {|a :+: b|} a (Inl x)   =  fresh a x
    fresh {|a :+: b|} a (Inr y)   =  fresh a y
    fresh {|a :*: b|} a (x :*: y) =  fresh a x && fresh a y

    aeq                                   :: a -> a -> Bool
    aeq {|Unit|}    Unit      Unit        =  True
    aeq {|a :+: b|} (Inl x)   (Inl x')    =  aeq x x'
    aeq {|a :+: b|} (Inr y)   (Inr y')    =  aeq y y'
    aeq {|a :+: b|} _         _           =  False
    aeq {|a :*: b|} (x :*: y) (x' :*: y') =  aeq x x' && aeq y y'

    perm      ::  Perm -> a -> a
    perm pi a =   foldr swap a pi

instance Nominal () where
    swap t x = x
    fresh a x = True
    aeq x y = x == y

instance Nominal Char where
    swap t x = x
    fresh a x = True
    aeq x y = x == y

instance Nominal Int where
    swap t x = x
    fresh a x = True
    aeq x y = x == y

instance Nominal Bool where
    swap t x  = x
    fresh a x = True
    aeq x y = x == y

instance Nominal Name where
    swap (a,b) c = if a == c then b 
		   else if b == c then a
		   else c
    fresh a b    = a /= b
    aeq x y = x == y

instance Nominal a => Nominal (Maybe a) where

instance (Nominal a, Nominal b) => Nominal (Either a b) where

instance (Nominal a) => Nominal [a] where

instance (Nominal a, Nominal b) => Nominal (a,b) where


-- Freshness monad type class.
-- A freshness monad is one that provides the ability to 
-- generate fresh names.

class Monad m => FreshMonad m where
    renameFM            :: Name -> m Name 

-- One example 
newtype StateFM x = FM (Int -> (x,Int))

instance Monad StateFM where
  return x = FM (\ n -> (x,n))
  (>>=) (FM h) g = FM f
    where f n = let (a,n2) = h n
                    FM k = g a
                in k n2

instance FreshMonad StateFM where
    renameFM (Name s _) = FM(\n -> (Name s (Just n),n+1))

runFM (FM f) = fst(f 0) 


