{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
{-# OPTIONS -fth #-}

{- Nominal library substitutions
   James Cheney
-}

module NominalSubst where

import SYBnew
import NominalBase
import List(union)

-- HasVar : class of types with a function var :: name -> a
-- and a test is_var :: a -> Maybe Name
-- If a type a has variables, then Subst a t and FreeVars a t
-- can be derived automatically.

class HasVar a where
    is_var :: a -> Maybe Name
    var :: Name -> a

instance HasVar Name where
    is_var a = Just a
    var a = a

class Subst t u where
    subst :: FreshMonad m => Name -> t -> u -> m u

data SubstD a t = SubstD {substD::FreshMonad m => Name -> a -> t -> m t}

instance Subst a t => Sat (SubstD a t) where
    dict = SubstD {substD = subst}

subst' :: (FreshMonad m,Data (SubstD a) t) => Name -> a -> t -> m t
subst' =  substD dict

substCtx :: SubstD a ()
substCtx =  undefined

instance Data (SubstD a) t => Subst a t where
    subst a t u = gmapM (substCtx :: SubstD a ())
		        (subst' a t) 
		        u

instance (HasVar a, Data (SubstD a) a) => Subst a a where
    subst a t x = if is_var x == Just a 
		  then return t 
		  else gfoldl (substCtx :: SubstD a ())
		         (\f x -> 
  		             do f' <- f;
		                t' <- subst' a t x;
		                return (f' t')) 
		         return 
			 x



class FreeVars a u where
    fvs :: a -> u -> [Name]


data FreeVarsD a t = FreeVarsD {fvsD::a -> t -> [Name]}

instance FreeVars a t => Sat (FreeVarsD a t) where
    dict = FreeVarsD {fvsD = fvs}


fvs' :: (Data (FreeVarsD a) t) => a -> t -> [Name]
fvs' =  fvsD dict

freeVarsCtx :: FreeVarsD a ()
freeVarsCtx =  undefined

newtype FVList a = FVList {unFVList::[Name]}

instance Data (FreeVarsD a) t => FreeVars a t where
    fvs dummy u = unFVList(
			  gfoldl (freeVarsCtx :: FreeVarsD a ())
				 (\(FVList fvs_f) x -> 
				     (FVList (List.union fvs_f (fvs' dummy x))))
				 (\_ -> FVList [])
				 u)





instance (HasVar a, Data (FreeVarsD a) a) => FreeVars a a where
   fvs dummy x = 
       case is_var x
       of Just a -> [a] 
	  Nothing -> unFVList (gfoldl (freeVarsCtx :: FreeVarsD a ())
			        (\(FVList fvs_f) x -> 
			           (FVList (List.union fvs_f (fvs' dummy x))))
                                (\_ -> FVList [])
			        x)


