{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
{-# OPTIONS -fth #-}
{-# OPTIONS -fgenerics #-}
module PolyLam where

import NominalBase
import NominalAbs
import NominalSubst
import SYBnew
import Derive
import List(find)

-- Second order lambda terms

data Type = VarTy Name
          | FnTy Type Type
          | AllTy (Abs Name Type) deriving Show

data Term = Var Name
          | App Term Term
          | Fn Type (Abs Name Term)
          | TyFn (Abs Name Term)
          | TyApp Term Type deriving Show

$(derive [''Type,''Term])

instance Nominal Type where
    -- generic

instance Nominal Term where
    -- generic

instance Eq Type where
    (==) = aeq

instance Eq Term where
    (==) = aeq

instance HasVar Type where
    var = VarTy
    is_var (VarTy x) = Just x
    is_var _ = Nothing

instance HasVar Term where
    var = Var
    is_var (Var x) = Just x
    is_var _ = Nothing

data TCEnv = TCEnv {tenv :: [Name], env :: [(Name,Type)]}
data TCM a = TCM ((TCEnv,Int) -> (a,Int))

instance Monad TCM where
  return x = TCM (\ (env,n) -> (x,n))
  (>>=) (TCM h) g = TCM f
    where f (env,n) = let (a,n2) = h (env,n)
                          TCM k = g a
                      in k (env,n2)

instance FreshMonad TCM where
    renameFM (Name s _) = TCM(\(env,next) -> (Name s (Just (next)),next +1))

runTCM (TCM f) = fst(f (TCEnv [] [], 0)) 


instance PolyTCM TCM where
    errorTC msg = error msg
    bindTV a (TCM m) = TCM(\(tc,n) -> m (TCEnv (a:tenv tc) (env tc),n))
    bindV a t (TCM m) = TCM(\(tc,n) -> m (TCEnv (tenv tc) ((a,t):env tc),n))
    lookupTV a  = TCM (\(tc,n) -> (a `elem` tenv tc ,n))
    lookupV a   = TCM (\(tc,n) -> 
		          case find (\(b,t) -> a == b) (env tc) 
			  of Just (a,t) -> (Just t,n)
			     Nothing -> (Nothing,n))

-- A freshness monad that also can store type and term variable bindings.
class FreshMonad m => PolyTCM m where
    bindTV   ::  Name -> m a -> m a
    bindV    ::  Name -> Type -> m a -> m a
    lookupTV ::  Name -> m Bool
    lookupV  ::  Name -> m (Maybe Type)
    errorTC  ::  String -> m a

wfTy                ::  PolyTCM m => Type -> m ()
wfTy (VarTy n)      =   do b <- lookupTV n
		           if b then return () else errorTC "Unbound type variable"
wfTy (FnTy t1 t2)   =   do wfTy t1;
		           wfTy t2
wfTy (AllTy abs)    =   do (a,ty) <- unAbs abs;
		           bindTV a (wfTy ty)

eqTy          ::  PolyTCM m => Type -> Type -> m ()
eqTy ty1 ty2  =   if ty1 == ty2 then return () else errorTC "type mismatch"

matchFnTy                   ::  PolyTCM m => Type -> m (Type,Type)
matchFnTy (FnTy ty1 ty2)    =   return (ty1,ty2)
matchFnTy _                 =   errorTC "Type mismatch: Expected function type"

matchAllTy              ::  PolyTCM m => Type -> m (Abs Name Type)
matchAllTy (AllTy abs)  =   return abs
matchAllTy _            =   errorTC "Type mismatch: Expected forall type"


inferTm                       ::  PolyTCM m => Term -> m Type
inferTm (Var x)               =   do  ty <- lookupV x
				      case ty 
				        of  Just ty' -> return ty'
				            Nothing -> errorTC "Unbound variable"
inferTm (App t1 t2)           =   do  ty1 <- inferTm t1
				      (ty1',ty2') <- matchFnTy ty1
				      ty2 <- inferTm t2
				      eqTy ty1' ty2
				      return ty2'
inferTm (Fn ty abs)           =   do  (a,t) <- unAbs abs
				      ty' <- bindV a ty (inferTm t)
				      return (FnTy ty ty')
				    
inferTm (TyApp tm ty)         =   do  ty' <- inferTm tm
				      wfTy ty
				      abs <- matchAllTy ty'
				      (a,ty'') <- unAbs abs
				      subst a ty ty''
inferTm (TyFn abs)            =   do  (a,tm) <- unAbs abs
				      ty <- bindTV a (inferTm tm)
				      return (AllTy (Abs a ty))
				    
		       

